import os
import shutil
import tempfile

import joblib
import pytest
from requests import HTTPError

from tests.tools import DF, DF2
from toucan_data_sdk.sdk import ToucanDataSdk, InvalidQueryError


def gen_client(mocker):
    class Response:
        content = 10

        def raise_for_status(self):
            pass

    resp = Response()
    client = mocker.MagicMock()
    client.sdk.post.return_value = resp
    return client


@pytest.fixture(name='sdk', scope='function')
def gen_data_sdk(mocker):
    sdk = ToucanDataSdk(
        instance_url='https://api-myinstance.toucantoco.com',
        small_app='demo',
        auth=('', '')
    )
    sdk.client = gen_client(mocker)
    yield sdk
    if os.path.exists(sdk.EXTRACTION_CACHE_PATH):
        shutil.rmtree(sdk.EXTRACTION_CACHE_PATH)


@pytest.fixture(name='sdk_old', scope='function')
def gen_data_sdk_old(mocker):
    sdk_old = ToucanDataSdk(
        instance_url='https://api-myinstance.toucantoco.com/demo',
        auth=('', '')
    )
    sdk_old.client = gen_client(mocker)
    yield sdk_old
    if os.path.exists(sdk_old.EXTRACTION_CACHE_PATH):
        shutil.rmtree(sdk_old.EXTRACTION_CACHE_PATH)


def gen_client_error(mocker):
    class Response:
        content = 10

        def raise_for_status(self):
            raise HTTPError()

    resp = Response()
    client = mocker.MagicMock()
    client.sdk.post.return_value = resp
    return client


@pytest.fixture(name='sdk_error', scope='function')
def gen_data_sdk_error(mocker):
    sdk = ToucanDataSdk('some_url', small_app='demo', auth=('', ''))
    sdk.client = gen_client_error(mocker)
    yield sdk
    if os.path.exists(sdk.EXTRACTION_CACHE_PATH):
        shutil.rmtree(sdk.EXTRACTION_CACHE_PATH)


@pytest.fixture(name='tmp_dir', scope='module')
def gen_tmp_dir():
    return tempfile.gettempdir()


@pytest.fixture(name='tmp_file', scope='function')
def gen_tmp_file(tmp_dir):
    tmp_file = tempfile.NamedTemporaryFile(dir=tmp_dir)
    yield tmp_file
    tmp_file.close()


@pytest.fixture(name='tmp_file2', scope='function')
def gen_tmp_file2(tmp_dir):
    tmp_file = tempfile.NamedTemporaryFile(dir=tmp_dir)
    yield tmp_file
    tmp_file.close()


def test_dfs(sdk, mocker):
    """It should use the cache properly"""
    mock_cache_exists = mocker.patch(
        'toucan_data_sdk.sdk.ToucanDataSdk.cache_exists')
    mock_read_cache = mocker.patch(
        'toucan_data_sdk.sdk.ToucanDataSdk.read_from_cache')
    mock_read_sdk = mocker.patch(
        'toucan_data_sdk.sdk.ToucanDataSdk.read_from_sdk')

    # 1. Cache directory exists
    mock_cache_exists.return_value = True
    mock_read_cache.return_value = {"domain_1": 1}
    assert sdk.get_dfs() == {"domain_1": 1}

    # 2. Cache directory does not exist
    mock_read_cache.reset_mock()
    mock_cache_exists.return_value = False
    mock_read_sdk.return_value = {"domain_2": 1}
    assert sdk.get_dfs() == {"domain_2": 1}
    mock_read_cache.assert_not_called()

    # 3. Cache directory exists with one domain
    mock_cache_exists.return_value = True
    mock_read_cache.return_value = {"domain_1": 1}
    assert sdk.get_dfs(['a']) == {"domain_1": 1}

    # 4. Cache directory doesn't exists with one domain
    mock_read_cache.reset_mock()
    mock_cache_exists.return_value = False
    mock_read_sdk.return_value = {"domain_2": 1}
    assert sdk.get_dfs(['a']) == {"domain_2": 1}
    mock_read_cache.assert_not_called()
    assert sdk.small_app_url == 'https://api-myinstance.toucantoco.com/demo'


def test_dfs_complex(sdk, mocker):
    """It should use the cache properly"""
    mock_cache_exists = mocker.patch(
        'toucan_data_sdk.sdk.ToucanDataSdk.cache_exists')
    mock_read_cache = mocker.patch(
        'toucan_data_sdk.sdk.ToucanDataSdk.read_from_cache')
    mock_read_sdk = mocker.patch(
        'toucan_data_sdk.sdk.ToucanDataSdk.read_from_sdk')

    mock_cache_exists.side_effect = [True, False]
    mock_read_cache.return_value = {"domain_1": 1}
    mock_read_sdk.return_value = {"domain_2": 1}
    assert sdk.get_dfs(['a', 'b']) == {"domain_1": 1, "domain_2": 1}


def test_dfs_http_error(sdk_error):
    """It should use the cache properly"""
    with pytest.raises(HTTPError):
        _ = sdk_error.get_dfs()


def test_read_from_cache(sdk):
    with tempfile.TemporaryDirectory() as tmp_dir:
        extraction_dir = os.path.join(tmp_dir, sdk.EXTRACTION_CACHE_PATH)
        sdk.EXTRACTION_CACHE_PATH = extraction_dir
        os.makedirs(extraction_dir)

        joblib.dump(DF, os.path.join(extraction_dir, 'a'))
        joblib.dump(DF2, os.path.join(extraction_dir, 'b'))

        dfs = sdk.read_from_cache()
        assert 'a' in dfs
        assert 'b' in dfs
        assert DF.equals(dfs['a'])
        assert DF2.equals(dfs['b'])

        dfs = sdk.read_from_cache(['a'])
        assert 'a' in dfs
        assert 'b' not in dfs
        assert DF.equals(dfs['a'])


def test_read_from_sdk(sdk, mocker):
    mock_write = mocker.patch('toucan_data_sdk.sdk.ToucanDataSdk.write')
    mock_write.return_value = {'df': DF, 'df2': DF2}
    dfs = sdk.read_from_sdk()
    assert dfs == {'df': DF, 'df2': DF2}


def test_write(sdk, mocker):
    mock_extract = mocker.patch('toucan_data_sdk.sdk.extract')
    mock_extract.return_value = {'a': DF, 'b': DF2}

    with tempfile.TemporaryDirectory() as tmp_dir:
        extraction_dir = os.path.join(tmp_dir, sdk.EXTRACTION_CACHE_PATH)
        sdk.EXTRACTION_CACHE_PATH = extraction_dir

        sdk.write({'a': DF, 'b': DF2})

        assert os.path.exists(extraction_dir)
        assert 'a' in os.listdir(extraction_dir)
        assert 'b' in os.listdir(extraction_dir)


def test_invalidate_cache(sdk):
    with tempfile.TemporaryDirectory() as tmp_dir:
        extraction_dir = os.path.join(tmp_dir, sdk.EXTRACTION_CACHE_PATH)
        sdk.EXTRACTION_CACHE_PATH = extraction_dir
        os.makedirs(extraction_dir)
        joblib.dump(DF, os.path.join(extraction_dir, 'a'))
        joblib.dump(DF2, os.path.join(extraction_dir, 'b'))

        assert sdk.cache_exists()
        sdk.invalidate_cache()
        assert not sdk.cache_exists()

        os.makedirs(extraction_dir)
        joblib.dump(DF, os.path.join(extraction_dir, 'a'))
        joblib.dump(DF2, os.path.join(extraction_dir, 'b'))

        assert sdk.cache_exists('a')
        sdk.invalidate_cache(['a'])
        assert not sdk.cache_exists('a')
        assert sdk.cache_exists('b')
        assert sdk.cache_exists()


def test_invalidate_cache_exception(sdk):
    with tempfile.TemporaryDirectory() as tmp_dir:
        sdk.invalidate_cache()

        extraction_dir = os.path.join(tmp_dir, sdk.EXTRACTION_CACHE_PATH)
        sdk.EXTRACTION_CACHE_PATH = extraction_dir
        os.makedirs(extraction_dir)
        joblib.dump(DF, os.path.join(extraction_dir, 'a'))

        assert sdk.cache_exists()
        sdk.invalidate_cache(['b', 'a'])
        assert not sdk.cache_exists('a')


def test_augment(sdk):
    sdk.client.config.augment.get().text = 'yo'
    assert sdk.get_augment() == 'yo'


def test_basemaps(sdk):
    sdk.client.basemaps.post().json.return_value = {'lala': 'lo'}
    assert sdk.query_basemaps({}) == {'lala': 'lo'}
    with pytest.raises(InvalidQueryError):
        sdk.query_basemaps('yo')


def test_sdk_compatibility(sdk_old, mocker):
    """It should use the cache properly"""
    mock_cache_exists = mocker.patch(
        'toucan_data_sdk.sdk.ToucanDataSdk.cache_exists')
    mock_read_cache = mocker.patch(
        'toucan_data_sdk.sdk.ToucanDataSdk.read_from_cache')
    mock_read_sdk = mocker.patch(
        'toucan_data_sdk.sdk.ToucanDataSdk.read_from_sdk')

    # 1. Cache directory exists
    mock_cache_exists.return_value = True
    mock_read_cache.return_value = {"domain_1": 1}
    assert sdk_old.get_dfs() == {"domain_1": 1}
    assert sdk_old.small_app_url == 'https://api-myinstance.toucantoco.com/demo'



import pandas as pd
import logging


def cumsum(df, new_column, column, index, date_column, date_format):
    """
    DEPRECATED : Creates a new column, which is the cumsum of the column
    :param df: the dataframe
    :param new_column: name of the new column
    :param column: name on which the cumulative sum is performed
    :param index: array of column names to keep as indices
    :param date_column: column name that represent the date
    :param date_format: format of the date
    :return:
    """
    logging.getLogger(__name__).warning(f"DEPRECATED: use compute_cumsum")
    date_temp = '__date_temp__'
    if isinstance(index, str):
        index = [index]
    levels = list(range(0, len(index)))
    df[date_temp] = pd.to_datetime(df[date_column], format=date_format)
    reference_cols = [date_temp, date_column]
    df = df.groupby(index + reference_cols).sum()
    df[new_column] = df.groupby(level=levels)[column].cumsum()
    df.reset_index(inplace=True)
    del df[date_temp]

    return df



import pandas as pd

from toucan_data_sdk.utils.helpers import (
    check_params_columns_duplicate,
    ParamsValueError
)


def add_missing_row(df, id_cols, reference_col, complete_index=None, method=None,
                    cols_to_keep=None):
    """
    Add missing row to a df base on a reference column
    - `id_cols` are the columns id to group,
    - `reference_col` is the column with groups missing values
    - `complete_index` (optional) a set of values used to add missing rows,
       by default use the function `unique` on reference_col. Can be dict for date_range
    - `method` (optional) method to choose values to keep.
       E.g between min and max value of the group.
    - `cols_to_keep` (optional) is the columns link to the reference_col to keep.

    For example :

    YEAR MONTH NAME  VALUE  X
    2017   1     A      1  lo
    2017   2     A      1  lo
    2017   3     A      1  la
    2017   1     B      1  la
    2017   3     B      1  la

    The function `add_missing_row` with the arguments :
            id_cols=['NAME']
            reference_col='MONTH'
    give as a result :


    YEAR MONTH NAME  VALUE  X
    2017   1     A      1  lo
    2017   2     A      1  lo
    2017   3     A      1  la
    2017   1     B      1  la
    2017   2     B      NA NA
    2017   3     B      1  la

    Args:
        df (pd.DataFrame):
        id_cols (list(str)):
        reference_col (str):
        complete_index (tuple/dict):
        method (str):
        keep_cols (list(str)):
    """
    if cols_to_keep is None:
        cols_for_index = [reference_col]
    else:
        cols_for_index = [reference_col] + cols_to_keep
    check_params_columns_duplicate(id_cols + cols_for_index)

    if method == 'between' or method == 'between_and_after':
        df['start'] = df.groupby(id_cols)[reference_col].transform(min)
        id_cols += ['start']
    if method == 'between' or method == 'between_and_before':
        df['end'] = df.groupby(id_cols)[reference_col].transform(max)
        id_cols += ['end']

    names = id_cols + cols_for_index
    new_df = df.set_index(names)
    index_values = df.groupby(id_cols).sum().index.values

    if complete_index is None:
        complete_index = df.groupby(cols_for_index).sum().index.values
    elif isinstance(complete_index, dict):
        if complete_index['type'] == 'date':
            freq = complete_index['freq']
            date_format = complete_index['format']
            start = complete_index['start']
            end = complete_index['end']
            if isinstance(freq, dict):
                freq = pd.DateOffset(**{k: int(v) for k, v in freq.items()})
            complete_index = pd.date_range(start=start, end=end, freq=freq)
            complete_index = complete_index.strftime(date_format)
        else:
            raise ParamsValueError(f'Unknown complete index type: '
                                   f'{complete_index["type"]}')

    if not isinstance(index_values[0], tuple):
        index_values = [(x,) for x in index_values]
    if not isinstance(complete_index[0], tuple):
        complete_index = [(x,) for x in complete_index]
    new_tuples_index = [x + y for x in index_values for y in complete_index]

    new_index = pd.MultiIndex.from_tuples(new_tuples_index, names=names)
    new_df = new_df.reindex(new_index).reset_index()

    if method == 'between' or method == 'between_and_after':
        new_df = new_df[new_df[reference_col] >= new_df['start']]
        del new_df['start']
    if method == 'between' or method == 'between_and_before':
        new_df = new_df[new_df[reference_col] <= new_df['end']]
        del new_df['end']

    return new_df



import pandas as pd


def roll_up(df, levels, groupby_vars, extra_groupby_cols=[],
            var_name='type', value_name='value', agg_func='sum'):
    """
    Move the hierarchy from the columns name to the rows (like a melt).
    Add higher level hierarchy information with pandas aggregation function.

    One DatFrame per level (all will be concatenated in the end), group by
    levels, apply aggregation function on groupby_vars. Add two extra columns:
    value_name and var_name, like a pandas melt.

    Args:
        df (DataFrame): DataFrame to work on...
        levels (list): Hierarchy. The order is important, from the top level
            to the lower level.
        groupby_vars (list): Columns to select from the group by (apply
            aggregation function to)
        extra_groupby_cols (list): Add to columns to group by each time.
        var_name (str): Same as a pandas melt() var_name
        value_name (str): Same as a pandas melt() value_name
        agg_func (str): pandas aggregation function to apply to the groupby.

    Returns:
        DataFrame:

    """
    dfs = list()
    groupby_cols_cpy = list(levels)
    levels_cpy = list(levels)
    levels_cpy.reverse()
    for top_level in levels_cpy:
        # Aggregation
        gb_df = getattr(
            df.groupby(groupby_cols_cpy + extra_groupby_cols)[groupby_vars],
            agg_func)().reset_index()

        # Melt-like columns
        gb_df[var_name] = top_level
        gb_df[value_name] = gb_df[top_level]
        dfs.append(gb_df)

        # Remove one level each time in the groupby: lowest level column needs
        # a groupby with every levels, the next level needs every one except
        # the lowest, etc. until the top level column that needs only itself
        # inside the groupby.
        groupby_cols_cpy.pop()
    return pd.concat(dfs).reset_index()



import logging
import os
import shutil
import tempfile
import zipfile

import joblib
from toucan_client import ToucanClient
from toucan_data_sdk.utils.helpers import slugify


logger = logging.getLogger(__name__)


class ToucanDataSdk:
    def __init__(self, instance_url, auth, small_app=None, stage="staging"):
        instance_url = instance_url.strip().rstrip('/')
        if small_app is None:
            small_app = instance_url.split('/')[-1]
            instance_url = '/'.join(instance_url.split('/')[:-1])
        self.small_app_url = instance_url + (('/' + small_app) if small_app else '')
        self.client = ToucanClient(self.small_app_url, auth=auth, stage=stage)
        self.EXTRACTION_CACHE_PATH = os.path.join(
            'extraction_cache',
            slugify(instance_url, delim='_'),
            small_app
        )

    def get_dfs(self, domains=None):
        if domains is not None and isinstance(domains, list):
            dfs = {}
            domains_cache = [domain for domain in domains
                             if self.cache_exists(domain)]
            domains_sdk = list(set(domains) - set(domains_cache))

            if len(domains_cache) > 0:
                dfs.update(self.read_from_cache(domains_cache))
            if len(domains_sdk) > 0:
                dfs.update(self.read_from_sdk(domains_sdk))
        else:
            if self.cache_exists():
                dfs = self.read_from_cache()
            else:
                dfs = self.read_from_sdk()

        return dfs

    def invalidate_cache(self, domains=None):
        if domains is not None and isinstance(domains, list):
            for domain in domains:
                try:
                    os.remove(os.path.join(self.EXTRACTION_CACHE_PATH, domain))
                except OSError:
                    pass
        else:
            try:
                shutil.rmtree(self.EXTRACTION_CACHE_PATH)
            except (OSError, IOError) as e:  # For Python 2.7+ compatibility
                logger.error('failed to remove cache for : ' + str(e))

    def get_augment(self):
        return self.client.config.augment.get().text

    def query_basemaps(self, query):
        if isinstance(query, dict):
            return self.client.basemaps.post(json=query).json()
        else:
            raise InvalidQueryError(f'Query {query} should be a dict, {type(query)} found.')

    def read_from_sdk(self, domains=None):
        # Extract all domains if domains_sdk is null
        resp = self.client.sdk.post(json={'domains': domains})
        resp.raise_for_status()
        dfs = self.write(resp.content)
        logger.info(f'Data {domains} fetched and cached')
        return dfs

    def read_from_cache(self, domains=None):
        """
        Returns:
            dict: Dict[str, DataFrame]
        """
        logger.info('Reading data from cache ({})'.format(self.EXTRACTION_CACHE_PATH))
        if domains is not None and isinstance(domains, list):
            dfs = {domain: self.read_entry(domain) for domain in domains}
        else:
            dfs = {name: self.read_entry(name)
                   for name in os.listdir(self.EXTRACTION_CACHE_PATH)}
        return dfs

    def read_entry(self, file_name):
        """
        Args:
            file_name (str):

        Returns:
            pd.DataFrame:
        """
        file_path = os.path.join(self.EXTRACTION_CACHE_PATH, file_name)
        logger.info('Reading cache entry: {}'.format(file_path))
        return joblib.load(file_path)

    def write(self, data):
        """
        Args:
            data (str | byte):

        Returns:
            dict: Dict[str, DataFrame]
        """
        if not os.path.exists(self.EXTRACTION_CACHE_PATH):
            os.makedirs(self.EXTRACTION_CACHE_PATH)

        dfs = extract(data)
        for name, df in dfs.items():
            self.write_entry(name, df)
        return dfs

    def write_entry(self, file_name, df):
        """
        Args:
            file_name (str):
            df (DataFrame):
        """
        file_path = os.path.join(self.EXTRACTION_CACHE_PATH, file_name)
        joblib.dump(df, filename=file_path)
        logger.info('Cache entry added: {}'.format(file_path))

    def cache_exists(self, domain=None):
        if domain is not None:
            path = os.path.join(self.EXTRACTION_CACHE_PATH, domain)
            return os.path.exists(path) and os.path.isfile(path)
        else:
            path = self.EXTRACTION_CACHE_PATH
            return os.path.exists(path) and os.path.isdir(path)


def extract_zip(zip_file_path):
    """
    Returns:
        dict: Dict[str, DataFrame]
    """
    dfs = {}
    with zipfile.ZipFile(zip_file_path, mode='r') as z_file:
        names = z_file.namelist()
        for name in names:
            content = z_file.read(name)
            _, tmp_file_path = tempfile.mkstemp()
            try:
                with open(tmp_file_path, 'wb') as tmp_file:
                    tmp_file.write(content)

                dfs[name] = joblib.load(tmp_file_path)
            finally:
                shutil.rmtree(tmp_file_path, ignore_errors=True)
    return dfs


def extract(data):
    """
    Args:
        data (str | byte):

    Returns:
        dict: Dict[str, DataFrame]

    """
    _, tmp_file_path = tempfile.mkstemp()
    try:
        with open(tmp_file_path, 'wb') as tmp_file:
            tmp_file.write(data)

        if zipfile.is_zipfile(tmp_file_path):
            return extract_zip(tmp_file_path)
        else:
            raise DataSdkError('Unsupported file type')
    finally:
        shutil.rmtree(tmp_file_path, ignore_errors=True)


class DataSdkError(Exception):
    pass


class InvalidQueryError(Exception):
    pass



import pandas as pd


def waterfall(df, date, value, start, end, upperGroup,  # noqa:C901
              insideGroup=None, breakdown=None):
    """
    Return a line for each bars of a waterfall chart, totals, groups, subgroups.
    Compute the variation and variation rate for each line.

    Args:
        upperGroup (dict)
            - id: name of the column that contains upperGroups unique IDs
            - label: not required, text displayed under each upperGroups bars,
                     using ID when it's absent
            - groupsOrder: not required, order of upperGroups
        insideGroup (dict)
            - id: name of the column that contains insideGroups unique IDs
            - label: not required, text displayed under each insideGroups bars,
                     using ID when it's absent
            - groupsOrder: not required, order of insideGroups
        date (str): name of the column that id the period of each lines
        value (str): name of the column that contains the vaue for each lines
        start (dict):
            - label: text displayed under the first master column
            - date: value in the date col that id lines for the first period
        end (dict):
            - label: text displayed under the last master column
            - date: value in the date col that id lines for the second period
        breakdown (list) : (not implemented)
        # fillValues (bool): (case when false is not implemented)

    TODO:
        * idea to implement breakdown, call waterfall() on data filtered by filter.
        * crash gracefully if data or config is not coherent
        * il y aura des cols en trop: les gérer
    """

    def rename_or_create(group, group_name, name, target, default):
        """
        Rename or create standard columns for optional options
        """
        col_name = name.format(group_name)
        if not group.get(target):
            df[col_name] = default
        else:
            df.rename(columns={group[target]: col_name}, inplace=True)

    def first(x):
        return x.iloc[0]

    def compute_totals(totals):
        """
        Compute two dataframes with value for start and end
        Args:
            totals(dataframe):

        Returns: Dataframe, Dataframe

        """
        result = {}
        time_dict = {'start': start, 'end': end}
        for time_name, time in time_dict.items():
            if not totals[totals['date'] == time['id']].empty:
                value = totals.loc[
                    totals['date'] == time['id'], 'value'
                ].values[0]
            else:
                value = 0
            result[time_name] = pd.DataFrame([{
                'value': value,
                'label': time['label']
            }])
        return result['start'], result['end']

    def merge_df(df):
        """
        Compute diff value between start and end
        Args:
            df(dataframe):

        Returns: Dataframe

        """
        start_df, end_df = df[df['date'] == start_id], df.loc[df['date'] == end_id]
        df = start_df.merge(end_df,
                            on=['insideGroup', 'upperGroup',
                                'upperGroup_order',
                                'insideGroup_order', 'upperGroup_label',
                                'insideGroup_label'],
                            how='outer',
                            suffixes=('_start', '_end'), )

        # necessary before calculating variation
        df[['value_start', 'value_end']] = df[['value_start', 'value_end']].fillna(0)
        df['value'] = df['value_end'] - df['value_start']
        df.drop(['date_start', 'date_end', 'value_end'], axis=1, inplace=True)
        df.rename(columns={'upperGroup': 'groups'}, inplace=True)
        return df

    def compute_inside_group(df):
        """
        Compute inside Group
        Args:
            df(dataframe):

        Returns: Dataframe

        """
        inside_group = df.copy()
        inside_group['type'] = 'child'
        inside_group['variation'] = inside_group['value'] / inside_group['value_start']
        inside_group.drop(['upperGroup_label', 'insideGroup', 'value_start'],
                          axis=1, inplace=True)
        inside_group.rename(columns={'insideGroup_label': 'label'}, inplace=True)
        return inside_group

    def compute_upper_group(df):
        """
        Compute upperGroup
        Args:
            df (Dataframe):

        Returns: Dataframe

        """
        upper_group = df.groupby(['groups']).agg({
            'value': sum,
            'value_start': sum,
            'upperGroup_label': first,
            'upperGroup_order': first
        }).reset_index()
        upper_group['type'] = 'parent'
        upper_group['variation'] = upper_group['value'] / upper_group['value_start']
        upper_group.drop(['value_start'], axis=1, inplace=True)
        upper_group.rename(columns={'upperGroup_label': 'label'}, inplace=True)
        return upper_group

    if len(df) == 0:
        return df

    start_id, end_id = start['id'], end['id']

    if breakdown is not None:
        raise NotImplementedError(
            'We will add breakdown support on your request, please contact the devs')
    if insideGroup is None:
        raise NotImplementedError(
            'We will add support for upperGroup only on you request, please contact the devs')

    # prepare the dataframe with standard column names

    df = df.rename(columns={
        upperGroup['id']: 'upperGroup',
        insideGroup['id']: 'insideGroup',
        date: 'date',
        value: 'value'
    })
    groups_dict = {'upperGroup': upperGroup, 'insideGroup': insideGroup}
    for g_name, g in groups_dict.items():
        rename_or_create(g, g_name, '{}_label', 'label', df[g_name])
        rename_or_create(g, g_name, '{}_order', 'groupsOrder', pd.np.nan)

    agg_conf = {'value': sum}
    agg_conf.update({'{}_label'.format(col): first for col in groups_dict.keys()})
    agg_conf.update({'{}_order'.format(col): first for col in groups_dict.keys()})
    df = df.groupby([*groups_dict.keys(), 'date']).agg(agg_conf).reset_index()
    start, end = compute_totals(df.groupby('date').agg({'value': sum}).reset_index())

    df = merge_df(df)

    groups = pd.concat([
        compute_upper_group(df),
        compute_inside_group(df)
    ]).sort_values(
        by=['upperGroup_order', 'groups', 'type', 'insideGroup_order', 'label'],
        ascending=[True, True, False, True, True])
    ret = pd.concat([start, groups, end])

    # We don't need specific sorting orders by groups now
    ret.loc[ret['type'] == 'child', 'order'] \
        = ret.loc[ret['type'] == 'child', 'insideGroup_order']
    ret.loc[ret['type'] == 'parent', 'order'] \
        = ret.loc[ret['type'] == 'parent', 'upperGroup_order']
    ret.drop(['insideGroup_order', 'upperGroup_order'], axis=1, inplace=True)

    return ret



def top(df, value, limit, order='asc', group=None):
    """
    Awesome method that achieves what NO query in any language can do: (DRUM ROLL)
    Get the top or flop N results based on a column value for each specified group columns
    Args:
        - group: String or array of strings for the columns,
                 on which you want to perform the group operation
        - value: String for the column name on which you will rank the results
        - order: String 'asc' or 'desc' to sort by ascending ou descending order
        - limit: Number to specify the N results you want to retrieve.
                 Use a positive number x to retrieve the first x results.
                 Use a negative number -x to retrieve the last x results.
    """
    ascending = order != 'desc'
    limit = int(limit)
    filter_func = 'nlargest' if (limit > 0) ^ ascending else 'nsmallest'

    def _top(df):
        return getattr(df, filter_func)(abs(limit), value).sort_values(by=value,
                                                                       ascending=ascending)

    if group is None:
        df = _top(df)
    else:
        df = df.groupby(group).apply(_top)

    return df



from .argmax import argmax
from .converter import convert_str_to_datetime, convert_datetime_to_str
from .cumsum import cumsum
from .fillna import fillna
from .math import multiply, divide, add, subtract
from .melt import melt
from .percentage import percentage
from .pivot import pivot, pivot_by_group
from .query_df import query, query_df
from .rank import rank
from .rename import rename
from .replace import replace
from .sort import sort
from .top import top
from .waterfall import waterfall



import inspect
import linecache
import re
import shutil
from pathlib import Path
from typing import List
from unicodedata import normalize

import joblib


def get_orig_function(f):
    """ Make use of the __wrapped__ attribute to find the original function
        of a decorated function. """
    try:
        while True:
            f = f.__wrapped__
    except AttributeError:
        return f


def get_param_value_from_func_call(param_name, func, call_args, call_kwargs):
    """
    Get the value of a function's parameter based on its signature
    and the call's args and kwargs.

    Example:
        >>> def foo(a, b, c=3, d=4):
        ...    pass
        ...
        >>> # what would be the value of "c" when calling foo(1, b=2, c=33) ?
        >>> get_param_value_from_func_call('c', foo, [1], {'b': 2, 'c': 33})
        33
    """
    signature = inspect.signature(func)
    params_list = signature.parameters.keys()
    if param_name not in params_list:
        raise TypeError(f"'{param_name}' not found in {func.__name__}"
                        f"parameters list ([{params_list}])")
    call = signature.bind(*call_args, **call_kwargs)
    call.apply_defaults()
    return call.arguments[param_name]


def get_func_sourcecode(func):
    """
    Try to get sourcecode using standard inspect.getsource().
    If the function comes from a module which has been created dynamically
    (not from the filesystem), then it tries to read the sourcecode on the
    filesystem anyway.
    /!\ can do weird things if the filesystem code slightly differs from
        the original module code.
    """

    def getsource(func):
        lines, lnum = getsourcelines(func)
        return ''.join(lines)

    def getsourcelines(func):
        lines, lnum = findsource(func)
        return inspect.getblock(lines[lnum:]), lnum + 1

    def findsource(func):
        file = getfile(func)  # file path
        module = inspect.getmodule(func, file)
        lines = linecache.getlines(file, module.__dict__)
        code = func.__code__
        lnum = code.co_firstlineno - 1
        pat = re.compile(r'^(\s*def\s)|(\s*async\s+def\s)|(.*(?<!\w)lambda(:|\s))|^(\s*@)')
        while lnum > 0:
            if pat.match(lines[lnum]):
                break
            lnum = lnum - 1  # pragma: no cover
        return lines, lnum

    def getfile(func):
        module = inspect.getmodule(func)
        return module.__file__

    try:
        return inspect.getsource(func)
    except Exception:
        return getsource(func)


def check_params_columns_duplicate(cols_name: List[str]) -> bool:
    params = [column for column in cols_name if column is not None]
    if len(set(params)) != len(params):
        duplicates = set([x for x in params if params.count(x) > 1])
        raise ParamsValueError(
            f'Duplicate declaration of column(s) {duplicates} in the parameters')
    else:
        return True


def slugify(text, delim='-'):
    """ Generate an ASCII-only slug """
    _punctuation_re = re.compile(r'[\t !"#$%&\':()*\-/<=>?@\[\\\]^_`{|},.\n]+')

    result = []
    for word in _punctuation_re.split(text.lower()):
        word = normalize('NFKD', word) \
            .encode('ascii', 'ignore') \
            .decode('utf-8')

        if word:
            result.append(word)

    return delim.join(result)


def resolve_dependencies(func_name, dependencies):
    """ Given a function name and a mapping of function dependencies,
        returns a list of *all* the dependencies for this function. """

    def _resolve_deps(func_name, func_deps):
        """ Append dependencies recursively to func_deps (accumulator) """
        if func_name in func_deps:
            return

        func_deps.append(func_name)
        for dep in dependencies.get(func_name, []):
            _resolve_deps(dep, func_deps)

    func_deps = []
    _resolve_deps(func_name, func_deps)
    return sorted(func_deps)


def clean_cachedir_old_entries(cachedir: str, func_name: str, limit: int) -> int:
    """Remove old entries from the cache"""
    if limit < 1:
        raise ValueError("'limit' must be greater or equal to 1")

    cache_entries = get_cachedir_entries(cachedir, func_name)
    cache_entries = sorted(cache_entries, key=lambda e: e.last_access, reverse=True)
    cache_entries_to_remove = cache_entries[limit:]
    for entry in cache_entries_to_remove:
        shutil.rmtree(entry.path, ignore_errors=True)

    return len(cache_entries_to_remove)


def get_cachedir_entries(cachedir: str, func_name: str) -> list:
    entries = joblib.memory._get_cache_items(cachedir)
    return [e for e in entries if Path(e.path).parent.name == func_name]


class ParamsValueError(Exception):
    """
    Exception raised when some parameters value are wrong
    """



import numpy as np


def rank(df, value_cols, group_cols=None, rank_cols_names=None, method='min', ascending=True):
    """
    This function creates rank columns based on numeric values to be ranked.

    :param df: the dataframe to which ranking column(s) will be added
    :param value_cols (str or list(str)): name(s) of the columns used
    :param group_cols (str or list(str), optionnal): name(s) of the column(s) used to
      create each group inside which independent ranking needs to be applied
    :param rank_cols_names (str or list(str), optionnal): the names of the added ranking columns.
      If not filled, the ranking will be named after the value_cols with a '_rank' suffix
    :param method (str, optional): method to use when encountering equal values:
        - 'min' (default): lowest rank in group
        - 'max': highest rank in group
        - 'average': average rank of group
        - 'first': ranks assigned in order the values appear in the series
        - 'dense': like 'min', but rank always increases by 1 between groups
    :param ascending (bool, optional): True (default) or False whether the rank should be
      determined based on ascending or descending order respectively
    :return: the df dataframe with the new ranking columns.


    Examples:

    Dataset df:

    ENTITY  YEAR  VALUE_1  VALUE_2
       A    2017    10       3
       A    2017    20       1
       A    2018    10       5
       A    2018    30       4
       B    2017    60       4
       B    2017    40       3
       B    2018    50       7
       B    2018    60       6

    1. rank(df, value_cols='VALUE_1') returns:

    ENTITY  YEAR  VALUE_1  VALUE_2  VALUE_1_rank
       A    2017    10       3           1
       A    2017    20       1           3
       A    2018    10       5           1
       A    2018    30       4           4
       B    2017    60       4           7
       B    2017    40       3           5
       B    2018    50       6           6
       B    2018    50       7           6

    2. rank(df, value_cols='VALUE_1', group_cols='YEAR', method='average') returns:

    ENTITY  YEAR  VALUE_1  VALUE_2  RANK_1
       A    2017    10       3.0     1.0
       A    2017    20       1.0     2.0
       A    2018    10       5.0     1.0
       A    2018    30       4.0     2.0
       B    2017    60       4.0     4.0
       B    2017    40       3.0     3.0
       B    2018    50       6.0     3.5
       B    2018    50       7.0     3.5

    3. rank(df, value_cols=['VALUE_1', 'VALUE_2'], group_cols=['ENTITY', 'YEAR'],
            rank_cols_name=['RANK_1', 'RANK_2']) returns:

    ENTITY  YEAR  VALUE_1  VALUE_2  RANK_1  RANK_2
       A    2017    10       3        1       2
       A    2017    20       1        2       1
       A    2018    10       5        1       2
       A    2018    30       4        2       1
       B    2017    60       4        2       2
       B    2017    40       3        1       1
       B    2018    50       6        1       1
       B    2018    50       7        1       2

    """

    value_cols = [value_cols] if not isinstance(value_cols, list) else value_cols
    for col in value_cols:
        if not np.issubdtype(df[col].dtype, np.number):
            raise TypeError(col + " specified in value_cols must be of numeric type")

    if rank_cols_names is None:
        rank_cols_names = [x + '_rank' for x in value_cols]

    if group_cols is None:
        df[rank_cols_names] = df[value_cols].rank(method=method, ascending=ascending)
    else:
        df[rank_cols_names] = (df.groupby(group_cols)[value_cols]
                                 .rank(method=method, ascending=ascending))

    if method != 'average':
        df[rank_cols_names] = df[rank_cols_names].astype('int')

    return df



import numpy as np
import pandas as pd


def pivot(df, index, column, value):
    """
    Pivot a dataframe. Reverse operation of melting. Useful for configuring evolution
    See pandas' pivot_table documentation for more details
    Args:
        - index (list): indexes argument of pd.pivot_table
        - column (str): column name to pivot on
        - value (str): column name containing the value to fill the pivoted df
    """
    if df.dtypes[value].type == np.object_:
        df = pd.pivot_table(df, index=index,
                            columns=column,
                            values=value,
                            aggfunc=lambda x: ' '.join(x))
    else:
        df = pd.pivot_table(df, index=index,
                            columns=column,
                            values=value)
    df = df.reset_index()
    return df


def pivot_by_group(df, variable, value, new_columns, groups, id_cols=None):
    if id_cols is None:
        index = [variable]
    else:
        index = [variable] + id_cols

    param = pd.DataFrame(groups, index=new_columns)
    temporary_colum = 'tmp'

    df[temporary_colum] = df[variable]
    for column in param.columns:
        df.loc[df[variable].isin(param[column]), variable] = column

    param = param.T
    for column in param.columns:
        df.loc[
            df[temporary_colum].isin(param[column]), temporary_colum] = column

    df = pivot(df, index, temporary_colum, value)
    return df



import pandas as pd


def convert_str_to_datetime(df, selector, format=None):
    """
    Convert string column into datetime column
    :param df: Dataframe
    :param selector: name of the column to format
    :param format: format of the values
    :return: df
    """
    df[selector] = pd.to_datetime(df[selector], format=format)
    return df


def convert_datetime_to_str(df, selector, format):
    """
    Convert datetime column into string column
    :param df: Dataframe
    :param selector: name of the column to format
    :param format: format of the result values
    :return: df
    """
    df[selector] = df[selector].dt.strftime(format)
    return df
